use crate::Context;
use axum::{
    body::{Body, Bytes},
    extract::State,
    http::{Response, StatusCode},
    response::IntoResponse,
    Json,
};
use futures_util::Stream;
use std::{
    convert::Infallible,
    pin::Pin,
    task::{Context as TaskContext, Poll},
    time::Duration,
};
use tokio::{
    sync::mpsc::{Receiver, Sender},
    time::{self, Interval},
};
use verax::Filter;

const DEFAULT_PING_INTERVAL_SECS: u64 = 30;

struct SubscriberStream {
    receiver: Receiver<verax::Transaction>,
    ping_interval: Interval,
    ping: u128,
}

impl SubscriberStream {
    pub fn new(subscription: (Sender<verax::Transaction>, Receiver<verax::Transaction>)) -> Self {
        Self {
            receiver: subscription.1,
            ping_interval: time::interval(Duration::from_secs(DEFAULT_PING_INTERVAL_SECS)),
            ping: 0,
        }
    }
}

impl Stream for SubscriberStream {
    type Item = Result<Bytes, Infallible>;

    fn poll_next(mut self: Pin<&mut Self>, cx: &mut TaskContext<'_>) -> Poll<Option<Self::Item>> {
        match Pin::new(&mut self.ping_interval).poll_tick(cx) {
            Poll::Ready(_) => {
                // Send a heartbeat message
                self.ping += 1;
                let message = format!("{}\"ping\":{}{}\n", "{", self.ping, "}");
                let heartbeat_bytes = Bytes::copy_from_slice(message.as_bytes());
                return Poll::Ready(Some(Ok(heartbeat_bytes)));
            }
            Poll::Pending => {}
        }

        match self.receiver.poll_recv(cx) {
            Poll::Ready(Some(msg)) => {
                let mut msg = serde_json::to_vec(&msg).unwrap_or_default();
                msg.push(b'\n');
                Poll::Ready(Some(Ok(Bytes::copy_from_slice(&msg))))
            }
            Poll::Ready(None) => Poll::Ready(None),
            Poll::Pending => Poll::Pending,
        }
    }
}

pub async fn handler(State(ctx): State<Context>, Json(tag): Json<Filter>) -> impl IntoResponse {
    Response::builder()
        .status(StatusCode::OK)
        .header("Content-Type", "application/json")
        .body(Body::from_stream(SubscriberStream::new(
            ctx.ledger.subscribe(tag).await,
        )))
        .unwrap_or_else(|_| StatusCode::INTERNAL_SERVER_ERROR.into_response())
}
