use crate::{Context, Handler};
use axum::{extract::State, http::StatusCode, Json};
use serde::Deserialize;
use verax::{AccountId, AnyAmount, ReplayProtection, Status};

#[derive(Deserialize)]
pub struct AccountWithAmount {
    pub account: AccountId,
    #[serde(flatten)]
    pub amount: AnyAmount,
}

#[derive(Deserialize)]
pub struct Transaction {
    pub debit: Vec<AccountWithAmount>,
    pub credit: Vec<AccountWithAmount>,
    pub memo: String,
    pub status: Status,
    pub replay_protection: Option<ReplayProtection>,
}

#[async_trait::async_trait]
impl Handler for Transaction {
    type Ok = verax::Transaction;
    type Err = verax::Error;

    async fn handle(self, ledger: &Context) -> Result<Self::Ok, Self::Err> {
        let from = self
            .debit
            .into_iter()
            .map(|x| x.amount.try_into().map(|amount| (x.account, amount)))
            .collect::<Result<Vec<_>, _>>()?;

        let to = self
            .credit
            .into_iter()
            .map(|x| x.amount.try_into().map(|amount| (x.account, amount)))
            .collect::<Result<Vec<_>, _>>()?;

        ledger
            .ledger
            .new_transaction(self.memo, self.status, from, to, self.replay_protection)
            .await
    }
}

pub async fn handler(
    State(ledger): State<Context>,
    Json(item): Json<Transaction>,
) -> Result<Json<verax::Transaction>, (StatusCode, String)> {
    item.handle(&ledger)
        .await
        .map_err(|err| {
            (
                StatusCode::BAD_REQUEST,
                serde_json::to_string(&err).unwrap_or_default(),
            )
        })
        .map(Json)
}
