use crate::{
    filter_expr::{
        expr::{Expr, Variable},
        Error,
    },
    payment::PaymentTo,
    AccountId, Amount, Asset, PaymentFrom, RevId, Status, Tag, TxId, Type,
};
use chrono::{DateTime, Utc};
use num::CheckedAdd;
use std::ops::{Add, Deref};

#[derive(Clone, Debug, PartialEq, PartialOrd)]
pub enum Value {
    /// Not a value
    Nil,
    /// Amount
    Amount(Amount),
    /// Asset
    Asset(Asset),
    /// Account Type
    AccountId(AccountId),
    /// Transaction ID
    TxId(TxId),
    /// Type
    Type(Type),
    /// Revision Id
    RevId(RevId),
    /// Status
    Status(Status),
    /// Tag
    Tag(Tag),
    /// Datetime
    Date(DateTime<Utc>),
    /// String
    String(String),
    /// Integer number
    Number(i128),
    /// PaymentFrom
    From(Vec<PaymentFrom>),
    /// Payment Recipient
    To(Vec<PaymentTo>),
    /// Tags
    Tags(Vec<Tag>),
    /// Bool values
    Bool(bool),
    /// A dynamic variable name
    Variable(Variable),
}

impl From<&Variable> for Value {
    fn from(value: &Variable) -> Self {
        Self::Variable(value.clone())
    }
}

impl From<bool> for Value {
    fn from(x: bool) -> Self {
        Value::Bool(x)
    }
}

impl From<i128> for Value {
    fn from(x: i128) -> Self {
        Value::Number(x)
    }
}

impl Value {
    /// Attempts to convert the value as a boolean if possible or returns an error.
    pub fn as_boolean(&self) -> Result<bool, Error> {
        match self {
            Value::Bool(b) => Ok(*b),
            Value::Nil => Ok(false),
            Value::Number(x) => Ok(*x > 0),
            _ => panic!("Cannot convert to bool"),
        }
    }

    pub fn as_number(&self) -> Result<i128, Error> {
        match self {
            Value::Number(x) => Ok(*x),
            Value::String(x) => x.parse().map_err(|_| Error::NotANumber),
            _ => Err(Error::NotANumber),
        }
    }
}

impl Add for Value {
    type Output = Self;

    fn add(self, other: Self) -> Self {
        Value::Number(
            self.as_number()
                .expect("Not a number")
                .checked_add(other.as_number().expect("Not a number"))
                .expect("Overflow"),
        )
    }
}

impl CheckedAdd for Value {
    fn checked_add(&self, other: &Self) -> Option<Self> {
        self.as_number()
            .ok()?
            .checked_add(other.as_number().ok()?)
            .map(Value::Number)
    }
}

#[derive(Debug, PartialOrd)]
/// Value or reference to a value.
///
/// A reference to a value is being used to avoid cloning from the source code to the registers,
/// instead to just use a readonly reference
pub enum ValueOrRef<'a> {
    Ref(&'a Value),
    Value(Value),
}

impl PartialEq for ValueOrRef<'_> {
    fn eq(&self, other: &Self) -> bool {
        match (self, other) {
            (ValueOrRef::Ref(a), ValueOrRef::Ref(b)) => a == b,
            (ValueOrRef::Value(a), ValueOrRef::Value(b)) => a == b,
            (ValueOrRef::Ref(a), ValueOrRef::Value(b)) => *a == b,
            (ValueOrRef::Value(a), ValueOrRef::Ref(b)) => a == *b,
        }
    }
}

impl<'a> Clone for ValueOrRef<'a> {
    fn clone(&self) -> Self {
        match self {
            ValueOrRef::Ref(value) => ValueOrRef::Value((*value).clone()),
            ValueOrRef::Value(value) => ValueOrRef::Value(value.clone()),
        }
    }
}

impl<'a> Into<Value> for ValueOrRef<'a> {
    fn into(self) -> Value {
        match self {
            ValueOrRef::Ref(value) => value.clone(),
            ValueOrRef::Value(value) => value,
        }
    }
}

impl Deref for ValueOrRef<'_> {
    type Target = Value;

    fn deref(&self) -> &Self::Target {
        match self {
            ValueOrRef::Ref(value) => *value,
            ValueOrRef::Value(value) => &value,
        }
    }
}

impl<'a> From<&'a Value> for ValueOrRef<'a> {
    fn from(value: &'a Value) -> Self {
        ValueOrRef::Ref(value)
    }
}

impl<'a> From<Value> for ValueOrRef<'a> {
    fn from(value: Value) -> Self {
        ValueOrRef::Value(value)
    }
}
